package edu.fiu.jtlex.algorithms;

import edu.fiu.jtlex.timeml.graph.ITimeMLGraph;
import edu.fiu.jtlex.timeml.graph.ITimeMLGraphTCSP;
import edu.fiu.jtlex.timeml.graph.TimeMLGraphTCSP;
import edu.fiu.jtlex.timeml.timeline.ITimeMLIndeterminacy;
import edu.fiu.jtlex.timeml.timeline.TimeMLIndeterminacy;

import org.jacop.core.IntVar;

import java.util.*;

/**
 * Default implementation of INewIndeterminacyDetector
 *
 * @since jTLEX 1.0
 */
public class IndeterminacyDetector implements IIndeterminacyDetector {

    private final IGraphTCSPSolver _solver;

    /**
     * Creates a new indeterminacy detector with {@link GraphTCSPSolver}
     * as default implementation.
     *
     * @since jTLEX 1.0
     */
    public IndeterminacyDetector() {
        this(new GraphTCSPSolver());
    }

    /**
     * Creates a new inconsistency detector with the given {@link IGraphTCSPSolver} implementation.
     *
     * @param solver An {@link IGraphTCSPSolver} implementation
     * @throws NullPointerException If the TCSP solver is null
     * @since jTLEX 1.0
     */
    public IndeterminacyDetector(IGraphTCSPSolver solver) {
        Objects.requireNonNull(solver);

        _solver = solver;
    }

    /**
     * Detects indeterminacies in the given TimeML graph.
     *
     * @param timeMLGraph The graph to analyze.
     * @return An {@link ITimeMLIndeterminacy} with the results, or null if there's no solution.
     * @throws NullPointerException If the TimeML graph is null
     * @since jTLEX 1.0
     */
    public ITimeMLIndeterminacy detect(ITimeMLGraph timeMLGraph) {

        Objects.requireNonNull(timeMLGraph);

        ITimeMLGraphTCSP tcsp = new TimeMLGraphTCSP(timeMLGraph);
        List<IntVar> shortestTimeLine = _solver.solve(tcsp);
        if (shortestTimeLine == null) {
            return null;
        }

        List<Boolean> indeterminatePairs = Arrays.asList(new Boolean[shortestTimeLine.size() - 1]);
        Collections.fill(indeterminatePairs, false);

        Set<Integer> indeterminateIndexes = new HashSet<>();

        //Create a hundred different solutions and store them in a list.
        List<List<IntVar>> timelines = new ArrayList<>();
        for (int i = 0; i < 100; i++) {
            List<IntVar> randomTimeLine = _solver.solveRandom(tcsp);
            if (randomTimeLine != null)
                timelines.add(randomTimeLine);
        }

        for (int left = 0, right = 1; right < indeterminatePairs.size(); left++, right++) {
            String leftHandSide = shortestTimeLine.get(left).id();
            String rightHandSide = shortestTimeLine.get(right).id();
            boolean checkRightHandSide = true;

            for (List<IntVar> timeline : timelines) {
                for (int first = 0, second = 1; second < timeline.size(); first++, second++) {
                    IntVar firstIv = timeline.get(first);
                    IntVar secondIv = timeline.get(second);
                    if (firstIv.id().equals(leftHandSide) && secondIv.id().equals(rightHandSide)) {
                        indeterminateIndexes.add(right);
                        indeterminatePairs.set(left, true);
                        checkRightHandSide = false;
                        break;
                    }
                }
                if (!checkRightHandSide) break;
            }
        }

        return new TimeMLIndeterminacy(shortestTimeLine, indeterminatePairs, indeterminateIndexes);
    }

}
