package edu.fiu.jtlex.algorithms;

import edu.fiu.jtlex.timeml.data.ITimeMLLink;
import edu.fiu.jtlex.timeml.graph.IPartitionResult;
import edu.fiu.jtlex.timeml.graph.ITimeMLGraph;
import edu.fiu.jtlex.timeml.graph.ITimeMLGraphTCSP;
import edu.fiu.jtlex.timeml.graph.ITimeMLNode;
import edu.fiu.jtlex.timeml.graph.TimeMLGraphTCSP;
import edu.fiu.jtlex.timeml.timeline.ITimeMLIndeterminacy;
import edu.fiu.jtlex.timeml.timeline.ITimeMLTimeline;
import edu.fiu.jtlex.timeml.timeline.TimeMLTimeline;

import org.jacop.core.IntVar;

import java.util.*;
import java.util.stream.Collectors;

/**
 * Default implementation of {@link ITLEX}.
 *
 * @see TLEXAlgorithm
 * @since jTLEX 1.0
 */
public class TLEX implements ITLEX {

    private final ITimeMLGraph _timeMLGraph;
    private final IInconsistencyDetector _inconsistencyDetector;
    private final IIndeterminacyDetector _indeterminacyDetector;
    private final IPartitioner _partitioner;
    private final IDisconnectivityProcessor _disconnectivityProcessor;
    private final IGraphTCSPSolver _tcspSolver;

    private ITimeMLIndeterminacy _indeterminacyResults;
    private IPartitionResult _partitionResult;
    private List<List<String>> _solutions;
    private ITimeMLTimeline _timeline;

	/**
	 * Creates a new TLEX with the specified information.
	 * 
	 * @param timeMLGraph The IGraph to analyze.
	 * @throws NullPointerException If the graph is null.
	 * @since jTLEX 1.0
	 */
    public TLEX(ITimeMLGraph timeMLGraph) {
        this(timeMLGraph, new InconsistencyDetector(), new IndeterminacyDetector(),
                new Partitioner(), new DisconnectivityProcessor(), new GraphTCSPSolver());
    }

	/**
	 * Creates a new TLEX with the specified information.
	 * 
	 * @param timeMLGraph an {@link ITimeMLGraph}, which is the graph to analyze.
	 * @param inconsistencyDetector	an {@link IInconsistencyDetector} to detect inconsistencies.
	 * @param indeterminacyDetector	an {@link IIndeterminacyDetector} to detect indeterminacies.
	 * @param partitioner an {@link IPartitioner} to partition the graph.
	 * @param disconnectivityProcessor an {@link IDisconnectivityProcessor} to find disconnectivities.
	 * @param tcspSolver an {@link IGraphTCSPSolver} to solve TCSPs.
	 * @throws NullPointerException If the graph or any of the dependencies are null.
	 * @since jTLEX 1.0
	 */
    public TLEX(ITimeMLGraph timeMLGraph, IInconsistencyDetector inconsistencyDetector,
                IIndeterminacyDetector indeterminacyDetector, IPartitioner partitioner,
                IDisconnectivityProcessor disconnectivityProcessor, IGraphTCSPSolver tcspSolver) {
		Objects.requireNonNull(timeMLGraph);
		Objects.requireNonNull(inconsistencyDetector);
		Objects.requireNonNull(indeterminacyDetector);
		Objects.requireNonNull(partitioner);
		Objects.requireNonNull(disconnectivityProcessor);
		Objects.requireNonNull(tcspSolver);

        _timeMLGraph = timeMLGraph;
        _inconsistencyDetector = inconsistencyDetector;
        _indeterminacyDetector = indeterminacyDetector;
        _partitioner = partitioner;
        _disconnectivityProcessor = disconnectivityProcessor;
        _tcspSolver = tcspSolver;
    }

	/**
	 * Gets a list of all the partitions.
	 * 
	 * @return The list of partitions.
	 * @since jTLEX 1.0
	 */
    @Override
    public List<ITimeMLGraph> getPartitions() {
        if (_partitionResult == null)
            _partitionResult = _partitioner.partitionGraph(_timeMLGraph);
        return _partitionResult.getPartitions();
    }

	/**
	 * Gets a list of the consistent partitions.
	 * 
	 * @return The list of consistent partitions.
	 * @since jTLEX 1.0
	 */
    @Override
    public List<ITimeMLGraph> getConsistentPartitions() {
        return getPartitions().stream()
                .filter(_inconsistencyDetector::isConsistent)
                .collect(Collectors.toList());
    }

	/**
	 * Gets a list of the inconsistent partitions
	 * 
	 * @return The list of inconsistent partitions
	 * @since jTLEX 1.0
	 */
    @Override
    public List<ITimeMLGraph> getInconsistentPartitions() {
        return getPartitions().stream()
                .filter(g -> !_inconsistencyDetector.isConsistent(g))
                .collect(Collectors.toList());
    }
    
	/**
	 * Gets the set of inconsistent subgraphs.
	 *
	 * This requires the graph to be partitioned, so if it hasn't been done already,
	 * this method as a consequence will partition the graph.
	 * 
	 * @return The set of inconsistent partitions
	 * @since jTLEX 1.0
	 */
    @Override
    public Set<ITimeMLGraph> getInconsistentSubGraphs() {
		if (_partitionResult == null)
			getPartitions();
		return _inconsistencyDetector.generateInconsistentSubgraphs(_timeMLGraph);
    }
    
	/**
	 * The list of solutions for the graph.
	 * 
	 * @return The list of string representing solutions.
	 * @since jTLEX 1.0
	 */
    @Override
    public List<List<String>> getSolutions() {
    	if (_solutions == null) {
            _solutions = new ArrayList<>();
	    	List<ITimeMLGraph> partitions = this.getPartitions();
	    	for (ITimeMLGraph partition: partitions) {
	    		if (_inconsistencyDetector.isConsistent(partition)) {
                    ITimeMLGraphTCSP graphTCSP = new TimeMLGraphTCSP(partition);
                    List<IntVar> solution = _tcspSolver.solve(graphTCSP);
                    if (solution != null)
	        		    _solutions.add(_tcspSolver.convertSolutionToString(solution));
	    		}
	    	}
	    }
    	return _solutions;
    }

	/**
	 * Returns the timeline of the graph.
	 * 
	 * @return an ITimeline object containing the timeline of the graph.
	 * @since jTLEX 1.0
	 **/
    @Override
    public ITimeMLTimeline getTimeline() {
		if (_timeline == null)
			_timeline = new TimeMLTimeline(getSolutions(), _partitionResult.getPartitionLinks());
        return _timeline;
    }
    
	/**
	 * Returns the results after checking for Indeterminacy
	 * 
	 * @return an IIndeterminacy Object containing the Indeterminacy results.
	 * @since jTLEX 1.0
	 */
    private ITimeMLIndeterminacy getIndeterminacyResults() {
        if (_indeterminacyResults == null)
            _indeterminacyResults = _indeterminacyDetector.detect(_timeMLGraph);
        return _indeterminacyResults;
    }

	/**
	 * Returns the length of the timeline.
	 * 
	 * @return the length of the main timeline, -1 if graph is inconsistent.
	 * @since jTLEX 1.0
	 */
    @Override
    public int getTimelineLength() {
        return getIndeterminacyResults().getTimelineLength();
    }

	/**
	 * Returns a list of Indeterminant Time Pairs.
	 * 
	 * @return a list of strings containing the Indeterminant Time Pairs.
	 * @since jTLEX 1.0
	 */
    @Override
    public List<String> getIndeterminantTimePairs() {
        return getIndeterminacyResults().indeterminantTimePairs();
    }

	/**
	 * Returns a percent value representing the indeterminacy score.
	 * 
	 * @return a double that is the percent value representing the indeterminacy score.
	 * @since jTLEX 1.0
	 */
    @Override
    public double getIndeterminacyScore() {
        return getIndeterminacyResults().getIndeterminacyScore();
    }

	/**
	 * Checks if the graph is consistent.
	 *
	 * This requires the graph to be partitioned, so if it hasn't been done already,
	 * this method as a consequence will partition the graph.
	 * 
	 * @return true if the graph is consistent, else false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean isConsistent() {
		if (_partitionResult == null)
			getPartitions();
        return _inconsistencyDetector.isConsistent(_timeMLGraph);
    }

	/**
	 * Produces a printable string representation of the object
	 *
	 * @return A printable string representation of the object.
	 * @since jTLEX 1.0
	 */
    @Override
    public String toString() {
		StringBuilder sb = new StringBuilder(2048);

		sb.append(_timeMLGraph);
		sb.append("Consistency: ").append(isConsistent() ? "Consistent" : "Inconsistent").append("\n\n");

		sb.append("PARTITIONS:").append("\n\n");

		sb.append("CONSISTENT PARTITIONS:").append("\n\n");
		for (ITimeMLGraph partition : getConsistentPartitions()) {
			sb.append(partition);
			ITLEX tlex = new TLEX(partition, _inconsistencyDetector, _indeterminacyDetector, _partitioner, _disconnectivityProcessor, _tcspSolver);
			sb.append("Indeterminacy Score: ");
			sb.append("%.4f".formatted(tlex.getIndeterminacyScore())).append("\n");
			List<String> mainTimeline = tlex.getTimeline().getMainTimeline();
			sb.append("Timeline Length: ").append(mainTimeline.size()).append("\n");
			sb.append("Timeline: ").append(mainTimeline).append("\n\n");
		}

		sb.append("INCONSISTENT PARTITIONS:").append("\n\n");
		for (ITimeMLGraph partition : getInconsistentPartitions()) {
			sb.append(partition).append("\n");
		}

		Set<ITimeMLGraph> inconsistentSubGraphs = _inconsistencyDetector.generateInconsistentSubgraphs(_timeMLGraph);
		sb.append("INCONSISTENT SUBGRAPHS:").append("\n");
		sb.append("Count: ").append(inconsistentSubGraphs.size()).append("\n\n");
		for (ITimeMLGraph subgraph : inconsistentSubGraphs) {
			sb.append(subgraph).append("\n");
		}

        return sb.toString();
    }

	/**
	 * Produces a JSON representation of the object.
	 *
	 * @return A JSON string representation of the object.
	 * @since jTLEX 1.0
	 */
	@Override
    public String toJson() {
		StringBuilder sb = new StringBuilder(2048);
		StringJoiner sj;

		sb.append("{");

		sb.append("\"nodes\":");
		sj = new StringJoiner(",", "[", "],");
		for (ITimeMLNode node : _timeMLGraph.getNodes()) {
			String json = node.toJson();
			sj.add(json);
		}
		sb.append(sj);

		sb.append("\"links\":");
		sj = new StringJoiner(",", "[", "],");
		for (ITimeMLLink link : _timeMLGraph.getLinks()) {
			String json = link.toJson();
			sj.add(json);
		}
		sb.append(sj);

		sb.append("\"partitions\":");
		sj = new StringJoiner(",", "[", "],");
		for (ITimeMLGraph partition : getPartitions()) {
			String json = partitionAsJson(partition);
			if (json != null) sj.add(json);
		}
		sb.append(sj);

		sb.append("\"isConsistent\":")
				.append(_inconsistencyDetector.isConsistent(_timeMLGraph))
				.append(",");

		sb.append("\"inconsistentSubGraphs\":");
		sj = new StringJoiner(",", "[", "]");
		Set<ITimeMLGraph> inconsistentSubGraphs = _inconsistencyDetector.generateInconsistentSubgraphs(_timeMLGraph);
		for (ITimeMLGraph graph : inconsistentSubGraphs) {
			String json = graph.toJson();
			sj.add(json);
		}
		sb.append(sj);

		sb.append("}");

        return sb.toString();
    }

	/**
	 * Produces a string JSON object for a partition
	 *
	 * @param partition The partition to serialize
	 * @return The serialized partition
	 * @since jTLEX 1.0
	 */
	private String partitionAsJson(ITimeMLGraph partition) {
		ITimeMLIndeterminacy indeterminacyResult = _indeterminacyDetector.detect(partition);
		if (indeterminacyResult == null)
			return null;

		StringBuilder sb = new StringBuilder(128);
		StringJoiner sj;

		sb.append("{");

		sb.append("\"nodeIds\":");
		sj = new StringJoiner(",", "[", "]");
		for (ITimeMLNode node : partition.getNodes()) {
			String json = "\"" + node.getIdStr() + "\"";
			sj.add(json);
		}
		sb.append(sj).append(",");

		sb.append("\"linkIds\":");
		sj = new StringJoiner(",", "[", "]");
		for (ITimeMLLink link : partition.getLinks()) {
			String json = "\"lid" + link.getLinkID() + "\"";
			sj.add(json);
		}
		sb.append(sj).append(",");

		ITLEX tlex = new TLEX(partition, _inconsistencyDetector, _indeterminacyDetector, _partitioner, _disconnectivityProcessor, _tcspSolver);
		sb.append("\"timeline\":").append(tlex.getTimeline().toJson()).append(",");

		sb.append("\"isConsistent\":")
				.append(_inconsistencyDetector.isConsistent(partition))
				.append(",");

		sb.append("\"indeterminantTimePairs\":");
		sj = new StringJoiner(",", "[", "]");
		List<String> timePairs = indeterminacyResult.indeterminantTimePairs();
		for (String timePair : timePairs) {
			String json = timePairAsJson(timePair);
			sj.add(json);
		}
		sb.append(sj);

		sb.append("}");

		return sb.toString();
	}

	/**
	 * Produces a JSON string for a timePair.
	 *
	 * @param timePair The timePair to serialize
	 * @return The serialized timePair
	 * @since jTLEX 1.0
	 */
	private String timePairAsJson(String timePair) {
		String id1 = timePair.substring(0, timePair.indexOf("&") - 2);
		String timePoint1 = timePair.substring(timePair.indexOf("&") - 1, timePair.indexOf("&"));
		String id2 = timePair.substring(timePair.indexOf("&") + 1, timePair.length() - 1);
		String timePoint2 = timePair.substring(timePair.length() - 1);

		return "{\"id1\":\"%s\",\"timePoint1\":\"%s\",\"id2\":\"%s\",\"timePoint2\":\"%s\"}"
				.formatted(id1, timePoint1, id2, timePoint2);
	}
}
