package edu.fiu.jtlex.algorithms;


import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Objects;

import edu.fiu.jtlex.timeml.graph.ITimeMLGraph;

/**
 * Implementation of {@link ITLEXAlgorithm}.
 * Represents the main entry point for jTLEX.
 *
 * @see ITLEXAlgorithm
 * @since jTLEX 1.0
 */
public class TLEXAlgorithm implements ITLEXAlgorithm {

	private final IWriter _writer;

	/**
	 * Creates a default instance that prints the results to console.
	 */
	public TLEXAlgorithm() {
		this(new ConsoleWriter());
	}

	/**
	 * Creates a TLEXAlgorithm instance with the given {@link IWriter}
	 * for displaying the results.
	 *
	 * @param writer An {@link IWriter} to display the results.
	 * @throws NullPointerException If writer is null
	 * @since jTLEX 1.0
	 */
	public TLEXAlgorithm(IWriter writer) {
		Objects.requireNonNull(writer);
		_writer = writer;
	}

	/**
	 * Receives a directory's path and processes all the timeML files in it (.tml)
	 * and uses the given {@link IWriter} during construction to display the results.
	 *
	 * @param sourceDirectory Directory to process.
	 * @throws NullPointerException If sourceDirectory is null
	 * @throws IllegalArgumentException If sourceDirectory is not a directory
	 * @since jTLEX 1.0
	 */
	@Override
	public void process(String sourceDirectory) {
		Objects.requireNonNull(sourceDirectory);

		File directory = new File(sourceDirectory);
		if (!directory.isDirectory())
			throw new IllegalArgumentException(sourceDirectory + " is not a directory.");

		String[] timeMLFilenames = directory.list((dir, name) -> name.endsWith(".tml"));
		Objects.requireNonNull(timeMLFilenames);



		try {

			for (int i = 0; i < timeMLFilenames.length; i++) {

				String timeMLFilename = timeMLFilenames[i];

				Path path = Paths.get(sourceDirectory, timeMLFilename);
				File timeMLFile = new File(path.toString());
				ITimeMLGraph graph = GraphReader.createTimeMLGraph(timeMLFile);
				ITLEX tlex = new TLEX(graph);

				_writer.write("TimeMLFile: " + timeMLFilename + "\n");
				_writer.write(tlex.toString());
				_writer.writeLine("=====================EOF=====================");

	
			}



		}
		catch (FileNotFoundException e) {
			System.out.println(e.getMessage());
		}
		catch (IOException e) {
			System.out.println("An IO error occurred:");
			System.out.println(e.getMessage());
		}
	}
}
