package edu.fiu.jtlex.timeml.data;

import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Set;


/**
 * Default implementation of {@code IText}.
 * @since jTLEX 1.0
 * @author arada002
 */
public class TimeMLText implements ITimeMLText {

    private final String rawText;
    private final String text;
    private final Set<ITimeMLTimex> timexes;
    private final Set<ITimeMLLink> links;
    private final Set<ITimeMLInstance> instances;
    private final Set<ITimeMLEvent> events;

    /**
     * Creates a new Text object.
     *
     * @param rawText   Full article text including the tags.
     *                  May not be <code>null</code>
     * @param text      Article text without the tags.
     *                  May not be <code>null</code>.
     * @param timexes   Map of timex objects.
     *                  May not be empty.
     * @param links     Set of link objects.
     *                  May not be empty.
     * @param instances Set of makeInstance objects.
     *                  May not be empty.
     * @param events 	Map of event objects.
     *                  May not be empty.
     * @throws NullPointerException     If rawText or text argument is <code>null</code>
     * @throws IllegalArgumentException If rawText or text argument is all whitespace or an empty string.
     *                                  If events, timexes, or links argument is an empty set.
     */
    public TimeMLText(String rawText, String text, final Map<Integer, ITimeMLTimex> timexes,
                final Set<ITimeMLLink> links, final Map<Integer, ITimeMLInstance> instances, final Map<Integer, ITimeMLEvent> events) {

        checkArguments(rawText, text, timexes, links, instances, events);

        this.rawText = rawText;
        this.text = text;
        Set<ITimeMLTimex> timexSet= new HashSet<ITimeMLTimex>(timexes.values());
        this.timexes = Set.copyOf(timexSet);
        this.links = Set.copyOf(links);

        Set<ITimeMLInstance> instanceSet= new HashSet<ITimeMLInstance>(instances.values());
        this.instances = Set.copyOf(instanceSet);

        Set<ITimeMLEvent> eventSet= new HashSet<ITimeMLEvent>(events.values());
        this.events = Set.copyOf(eventSet);
    }

    /**
     * Returns the text.
     *
     * @return the String containing text.
     * @since jTLEX 1.0
     */
    public String getText() {
        return text;
    }

    /**
     * Returns the raw text.
     *
     * @return the String containing raw text.
     * @since jTLEX 1.0
     */
    public String getRawText() {
        return rawText;
    }

    /**
     * Returns the Timex's.
     *
     * @return the set of ITimeMLTimex's.
     * @since jTLEX 1.0
     */
    public Set<ITimeMLTimex> getTimexes() {
        return timexes;
    }

    /**
     * Returns the links.
     *
     * @return the set of ITimeMLLinks.
     * @since jTLEX 1.0
     */
    public Set<ITimeMLLink> getLinks() {
        return links;
    }

    /**
     * Returns the instances.
     *
     * @return the set of ITimeMLInstances.
     * @since jTLEX 1.0
     */
    public Set<ITimeMLInstance> getInstances() {
        return instances;
    }
    
    /**
     * Returns the instances.
     *
     * @return the set of ITimeMLEvents.
     * @since jTLEX 1.0
     */
    public Set<ITimeMLEvent> getEvents() {
        return events;
    }

    /**
     * Finds and returns the related instances to an event.
     * May be null.
     * @return a set of IInstances that are related to a given event.
     * @since jTLEX 1.0
     */
    public Set<ITimeMLInstance> getRelatedInstances(ITimeMLEvent event) {
        Set<ITimeMLInstance> relatedInstances = new HashSet<ITimeMLInstance>();
        for (ITimeMLInstance instance : instances) {
            if (event.equals(instance.getEvent())) {
                relatedInstances.add(instance);
            }
        }
        return relatedInstances;
    }

    /**
     * Utility method that checks if arguments passed to a constructor are valid.
     *
     * @param rawText   text with the tags
     * @param text      text without tags
     * @param timexes   set of timexes
     * @param links     set of links
     * @param instances set of event instances
     * 
     * @since jTLEX 1.0
     * 
     * @throws NullPointerException     If rawText or text argument is <code>null</code>
     * @throws IllegalArgumentException If rawText or text argument is all whitespace or an empty string.
     *                                  If events, timexes, or links argument is an empty set.
     */
    private void checkArguments(
            String rawText, String text, Map<Integer, ITimeMLTimex> timexes, Set<ITimeMLLink> links, Map<Integer, ITimeMLInstance> instances, Map<Integer, ITimeMLEvent> events) {
        if (rawText == null) {
            throw new NullPointerException("Raw text can not be null");
        }
        rawText = rawText.trim();
        if (rawText.length() == 0) {
            throw new IllegalArgumentException("Raw text can not be empty or all whitespace");
        }
        if (text == null) {
            throw new NullPointerException("Text can not be null");
        }
        text = text.trim();
        if (text.length() == 0) {
            throw new IllegalArgumentException("Text can not be empty or all whitespace");
        }
        if (timexes.isEmpty()) {
            throw new IllegalArgumentException("No timex objects found");
        }
        if (links.isEmpty()) {
            throw new IllegalArgumentException("No link objects found");
        }
        if (instances.isEmpty()) {
            throw new IllegalArgumentException("Missing makeInstances");
        }
        if (events.isEmpty()) {
            throw new IllegalArgumentException("No event objects found");
        }
    }
    
    /**
	 * Generates a hash code for the text.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
	 */
    @Override
    public int hashCode() {
        int hashCode = 1;;
        hashCode = 31*hashCode + (this.rawText==null ? 0 : this.rawText.hashCode());

        return hashCode;
    }
	/**
	 * Checks if this text equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both texts are equal, i.e. same rawText, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o) {
        Objects.requireNonNull(o);

        if (!(o instanceof ITimeMLText))
            return false;

        ITimeMLText obj = (TimeMLText) o;

        if (this.rawText != obj.getRawText())
            return false;

        return true;
    }
}
