package edu.fiu.jtlex.timeml.data;

import java.util.Objects;

/**
 * Default implementation of {@code ITimeMLTimex}.
 * @since jTLEX 1.0
 * @author ceber003
 */
public class TimeMLTimex implements ITimeMLTimex {

    private final int tID;
    private final TimexType type;
    private final String value;
    private final TimexMod mod;
    private final boolean temporalFunction;
    private final int anchorID;
    private final FunctionInDocument documentFunction;
    private final int beginPoint;
    private final int endPoint;
    private final String quant;
    private final String freq;
    private final String phrase;

    /**
     * Creates a new Timex with the specified information.
     *
     * @param tID              id of the timex; may not be negative
     * @param type             the type of the timex may not be null
     * @param value            the value attribute of the timex
     * @param mod              the mod of the value, is optional
     * @param temporalFunction true or false
     * @param anchorID         the id of another timex anchored to the current
     *                         timex, optional
     * @param documentFunction the function a timex is providing within the
     *                         document, is 'NONE' when not specified.
     * @param beginPoint       beginpoint is used if the timex is anchored to
     *                         another, can be null
     * @param endPoint         end is used if timex is anchored to another timex,
     *                         can be null
     * @param quant            used only if timex type is 'SET' is the literal
     *                         describing the timex rate of occurrence
     * @param freq             used only if timex type is 'SET' is the frequency the
     *                         timex occurs
     * @param phrase           the phase associated with this this timex.
     * @throws NullPointerException     if type or value is null
     * @throws IllegalArgumentException if tID, anchorId, beginPoint, or endPoint is
     *                                  negative
     */
    public TimeMLTimex(int tID, TimexType type, String value, TimexMod mod, boolean temporalFunction, int anchorID,
                 FunctionInDocument documentFunction, int beginPoint, int endPoint, String quant, String freq,
                 String phrase) {

        if (tID < 0) {
            throw new IllegalArgumentException("Error: Time id can't be less than zero.");
        }
        if (type == null) {
            throw new NullPointerException("Error: Timex type can't be Null.");
        }
        if (value == null) {
            throw new NullPointerException("Error: Timex value can't be Null.");
        }
        if (type == ITimeMLTimex.TimexType.DATE && getValueDigits(value) != true) {
            System.out.println("Error: Timex value is incorrect. Expected at least 4 digits. Please fix the value for t" + tID);
        }
        if (anchorID < 0) {
            throw new IllegalArgumentException("Error: Anchor id can't be less than zero.");
        }
        if (beginPoint < 0) {
            throw new IllegalArgumentException("Error: beginPoint id can't be less than zero.");
        }
        if (endPoint < 0) {
            throw new IllegalArgumentException("Error: endPoint id can't be less than zero.");
        }

        this.tID = tID;
        this.type = type;
        this.value = value;
        this.mod = mod;
        this.temporalFunction = temporalFunction;
        this.anchorID = anchorID;
        this.documentFunction = documentFunction;
        this.beginPoint = beginPoint;
        this.endPoint = endPoint;
        this.quant = quant;
        this.freq = freq;
        this.phrase = phrase;

        if (documentFunction == null) {
            documentFunction = FunctionInDocument.NONE;
        }
    }

    /**
     * Returns true if the the value is at least 4 digits.
     * Also checks if value equals PXY, PAST_REF, PRESENT_REF or FUTURE_REF to return true.
     * Will return false if null.
     * @since jTLEX 1.0
     * @param value a String that contains a numeric value
     * @return true if the value has at least 4 digits
     */
    public boolean getValueDigits(String value) {
        int digits = 0;
        String value2 = value.replaceAll("\\s+", "");
        if (value2.equals("PXY") || value2.equals("PAST_REF") || value2.equals("PRESENT_REF") || value2.equals("FUTURE_REF")) {
            return true;
        }

        for (int i = 0; i < value.length(); i++) {
            if (value.charAt(i) != ' ') {
                digits++;
            }
        }
        if (digits < 4) {
            return false;
        } else return true;

    }

    /**
     * Returns the tID of the timex.
     * Will never be null as it is required for the constructor.
     * @since jTLEX 1.0
     * @return the tID of the timex.
     */
    public int getId() {
        return tID;
    }

    /**
     * Grabs the tID, adds a t to the front for String format and returns.
     * Will never be null since the tID was required in the constructor.
     * @since jTLEX 1.0
     * @return the tID of the timex in String format.
     */
    public String getIdStr() {
        String tId = "t" + this.tID;
        return tId;
    }

    /**
     * Returns the type of the timex.
     * @since jTLEX 1.0
     * @return the type of the timex.
     */
    public TimexType getType() {
        return type;
    }

    /**
     * Returns the string representation of the value.
     * The value of the the timex tag determines what the actual amount of time for the
     * date, time, duration, and set types.
     * Will never be null as it is required in the constructor.
     * @since jTLEX 1.0
     * @return the string representation of the value.
     */
    public String getValue() {
        return value;
    }

    /**
     * Returns the mod to which the object belongs.
     * This value is optional and can be null.
     * @since jTLEX 1.0
     * @return the mod to which the object belongs.
     * @see TimexMod
     */
    public TimexMod getMod() {
        return mod;
    }

    /**
     * Returns the inary attribute which expresses that the value of
     * the temporal expression needs to be determined via evaluation of a temporal
     * function.
     * Cannot be null as is a required boolean.
     * @since jTLEX 1.0
     * @return the inary attribute which expresses that the value of
     * the temporal expression needs to be determined via evaluation of a temporal
     * function.
     */
    public boolean getTemporalFunction() {
        return temporalFunction;
    }

    /**
     * Returns the integer part of the anchortimeId.
     * Each anchortime has to be identified by a unique ID number.
     * This value is optional. If included, it must not be negative.
     * @since jTLEX 1.0
     * @return the integer part of the anchortimeId.
     */
    public int getAnchorTimeId() {
        return anchorID;
    }

    /**
     * Returns the string part of the anchortimeId.
     * Each anchortime has to be identified by a unique ID number.
     * This value is optional. If included, it must not be negative.
     * @since jTLEX 1.0
     * @return the string part of the anchortimeId.
     * May not be null.
     */
    public String getAnchorTimeIdStr() {
        String anchorId = "t" + anchorID;
        return anchorId;
    }

    // public String getValueFromFunction();

    /**
     * Returns the function to which the object belongs.
     * Function in document can be one of seven values.
     * If it is not specified then it will be set as 'NONE' by default.
     * It can be set to null.
     * @since jTLEX 1.0 
     * @return the function to which the object belongs.
     */
    public FunctionInDocument getFunction() {
        return documentFunction;
    }

    /**
     * Returns the integer part of the beginpointId.
     * Beginpoint is only used if the expression is anchored to another timex expression.
     * Each beginpoint has to be identified by a unique ID number.
     * If only endpoint is provided may create an empty timex to fill in.
     * It cannot be null, as it is required in constructor. It cannot be negative either.
     * @since jTLEX 1.0
     * @return the integer part of the beginpointId.
     * May not be negative
     */
    public int getBeginPoint() {
        return beginPoint;
    }

    /**
     * Returns the string part of the beginpointId.
     * Beginpoint is only used if the expression is anchored to another timex expression.
     * Each beginpoint has to be identified by a unique ID number.
     * If only endpoint is provided may create an empty timex to fill in.
     * It cannot be null as it is required in constructor. It cannot be negative either.
     * @since jTLEX 1.0
     * @return the string part of the beginpointId.
     */
    public String getBeginPointStr() {
        String beginPointId = "t" + beginPoint;
        return beginPointId;
    }

    /**
     * Returns the integer part of the endpointId.
     * Endpoint is only used if the expression is anchored to another timex expression.
     * Each endpoint has to be identified by a unique ID number.
     * If only beginpoint is provided may create an empty timex to fill in.
     * It cannot be null as it is required in constructor. It cannot be negative either.
     * @since jTLEX 1.0
     * @return the integer part of the endpointId.
     */
    public int getEndPoint() {
        return endPoint;
    }

    /**
     * Returns the string part of the endpointId.
     * Endpoint is only used if the expression is anchored to another timex expression.
     * Each endpoint has to be identified by a unique ID number.
     * If only beginpoint is provided may create an empty timex to fill in.
     * It cannot be null as it is required in constructor. It cannot be negative either.
     * @since jTLEX 1.0
     * @return the string part of the endpointId.
     */
    public String getEndPointStr() {
        String endPointId = "t" + endPoint;
        return endPointId;
    }

    /**
     * Returns the quantity to which the object belongs.
     * The quant value is only used if the timex type is 'SET'.
     * It can be set to null.
     * @since jTLEX 1.0
     * @return the quantity to which the object belongs.
     */
    public String getQuant() {
        return quant;
    }

    /**
     * Returns frequency String portion.
     * The frequency value is only used if the timex type is 'SET'.
     * The integer value determines the number of time the frequency takes.
     * i.e: 2 Weeks, 3 Months, 1 Day, etc.
     * It can be set to null.
     * @since jTLEX 1.0
     * @return frequency String portion.
     */
    public String getFreq() {
        return freq;
    }

    /**
     * Returns the phrase from the text annotated as timex.
     * It can be set to null.
     * @since jTLEX 1.0
     * @return the phrase from the text annotated as timex.
     */
    public String getPhrase() {
        return phrase;
    }


    /**
     * Returns the info of the Timex in String format.
     * Will never return null.
     * @since jTLEX 1.0
     * @return the info of the Timex in String format.
     */
    @Override
    public String toString() {
        String ret = "Timex: {tID = " + this.getIdStr() +
                ", Type = " + this.getType() +
                ", Value = " + this.getValue() +
                ", Mod = " + this.getMod() +
                ", Temporal Function? " + this.getTemporalFunction() +
                ", AnchorID = " + this.getAnchorTimeIdStr() +
                ", Begin Point = " + this.getBeginPointStr() +
                ", End Point = " + this.getEndPointStr() +
                ", Quantity = " + this.getQuant() +
                ", Frequency = " + this.getFreq();

        ret += "}";
        return ret;
    }

    /**
     * Converts the Timex and its info into JSON format.
     * Will never return null.
     * @since jTLEX 1.0
     * @return the JSON format of the Timex.
     */
    public String toJson() {
        String ret = "{\"id\":\"" + this.getIdStr() +
                "\", \"type\":\"" + this.getType().toString() +
                "\", \"value\":\"" + this.getValue();

        if (this.getMod() != null)
            ret += "\", \"mod\":\"" + this.getMod().toString();

        ret += "\", \"temporalFunction\":\"" + this.getTemporalFunction() +

                "\", \"anchorID\":\"" + this.getAnchorTimeIdStr() +
                "\", \"beginPoint\":\"" + this.getBeginPointStr() +
                "\", \"endPoint\":\"" + this.getEndPointStr() +
                "\", \"quantity\":\"" + this.getQuant() +
                "\", \"frequency\":\"" + this.getFreq() + "\"";

        ret += "}";
        return ret;
    }

    /**
     * Creates and returns a field-by-field copy of the Timex.
     * Will never return null.
     * @since jTLEX 1.0
     * @return a copy of the Timex.
     */
    public TimeMLTimex clone() {
        return new TimeMLTimex(
                this.getId(),
                (ITimeMLTimex.TimexType) this.getType(),
                this.getValue(),
                (ITimeMLTimex.TimexMod) this.getMod(),
                this.getTemporalFunction(),
                this.getAnchorTimeId(),
                (ITimeMLTimex.FunctionInDocument) this.getFunction(),
                this.getBeginPoint(),
                this.getEndPoint(),
                this.getQuant(),
                this.getFreq(),
                this.getPhrase());
    }
    
    
	/**
	 * Checks if this Timex equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both timexes are equal, i.e. same ID and type, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o) {
        Objects.requireNonNull(o);

        if (!(o instanceof ITimeMLTimex))
            return false;

        ITimeMLTimex obj = (TimeMLTimex) o;

        if (this.tID != obj.getId())
            return false;
        
        if (this.type != obj.getType())
        	return false;

        if (this.value != obj.getValue())
        	return false;
        
        if (this.mod != obj.getMod())
        	return false;
        
        if (this.temporalFunction != obj.getTemporalFunction())
        	return false;
        
        if (this.anchorID != obj.getAnchorTimeId())
        	return false;
        
        if (this.documentFunction != obj.getFunction())
        	return false;
        
        if (this.beginPoint != obj.getBeginPoint())
        	return false;
        
        if (this.endPoint != obj.getEndPoint())
        	return false;
        
        if (this.quant != obj.getQuant())
        	return false;
        
       if (this.freq != obj.getFreq())
    	   return false;
       
       if (this.phrase != obj.getPhrase())
    	   return false;

        return true;
    }
}
