package edu.fiu.jtlex.timeml.timeline;

import java.util.Iterator;
import java.util.List;

/**
 * Represents a Timeline of a TimeML graph, which is an ordered list of the events
 * extracted from a TimeML Graph. TimeML annotations are converted into a collection of
 * main and subordinated timelines arranged into a trunk-and-branch style structure.
 * This class takes an ordered timeline, a set of attachment links, arranging them into a Timeline.
 * @since jTLEX 1.0
 */
public interface ITimeMLTimeline {

	/**
	 * Returns an iterator object that is used in for-each loops.
	 * @return an iterator object that is used in for-each loops.
	 * @since jTLEX 1.0
	 */
	Iterator<String> iterator();
	
	
	/**
	 * Checks if there are Timeline elements remaining in the list.
	 * @return true if there are Timeline elements remaining in the list.
	 * @since jTLEX 1.0
	 */
	boolean hasNext();

	/**
	 * Returns next String representation of an event on the timeline.
	 * @return next String representation of an event on the timeline.
	 * @since jTLEX 1.0
	 */
	String next();
	
	/**
	 * Returns the {@code List<String>} representation of the main Timeline in the graph.
	 * @return the {@code List<String>} representation of the main Timeline in the graph.
	 * @since jTLEX 1.0
	 */
	List<String> getMainTimeline();
	
	/**
	 * Returns an ArrayList containing the Timelines belonging to the subordinating subgraphs.
	 * @return an ArrayList containing the Timelines belonging to the subordinating subgraphs.
	 * @since jTLEX 1.0
	 */
	List<List<String>> getSubordinationTimelines();
	
	/**
	 * Returns number of timepoints.
	 * @return the number of timepoints.
	 * @since jTLEX 1.0
	 */
	int getTotalTimePoints();
	
	/**
	 * Returns a List containing the breaking points from the Partitioning.java class.
	 * @return a List containing the breaking points from the Partitioning.java class.
	 * @since jTLEX 1.0
	 */
	List<String> getAttachmentPoints();
	
	/**
	 * Returns the first timepoint in the Timeline.
	 * @return the first timepoint in the Timeline.
	 * @since jTLEX 1.0
	 */
	String getFirstPoint();
	
	/**
	 * Returns the last timepoint in the Timeline.
	 * @return the last timepoint in the Timeline.
	 * @since jTLEX 1.0
	 */
	String getLastPoint();
	
	/**
	 * Returns the number of events in the timeline.
	 * @return the number of events in the timeline.
	 * @since jTLEX 1.0
	 */
	int getTimelineLength();
	
	/**
	 * Returns the total number of subordination branches in the graph.
	 * @return the total number of subordination branches in the graph.
	 * @since jTLEX 1.0
	 */
	int getTotalSubordinationBranches();

	
	/**
	 * Returns the JSON (RFC 8259) representation of the Timeline.
	 * @return the JSON (RFC 8259) representation of the Timeline.
	 * @since jTLEX 1.0
	 */
	String toJson();
	
	/**
	 * Generates a hash code for the timeline.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
	 */
    @Override
    public int hashCode();
	
	/**
	 * Checks if this Timeline equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if timelines are equal, i.e. same mainTimeline and timelines, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o);
}
